package io.testcase;


import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.util.concurrent.AtomicDouble;
import io.testcase.model.Access;
import io.testcase.model.PayLoadUsable;
import io.testcase.model.Satellite;
import io.testcase.model.Target;
import net.iakovlev.timeshape.TimeZoneEngine;
import org.geojson.Feature;
import org.geojson.FeatureCollection;
import org.geojson.LngLatAlt;
import org.geojson.Polygon;
import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.geometry.spherical.twod.SphericalPolygonsSet;
import org.hipparchus.ode.events.Action;
import org.hipparchus.util.FastMath;
import org.orekit.attitudes.AttitudeProvider;
import org.orekit.attitudes.NadirPointing;
import org.orekit.bodies.*;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.TopocentricFrame;
import org.orekit.geometry.fov.DoubleDihedraFieldOfView;
import org.orekit.models.earth.tessellation.EllipsoidTessellator;
import org.orekit.orbits.KeplerianOrbit;
import org.orekit.orbits.Orbit;
import org.orekit.orbits.PositionAngle;
import org.orekit.propagation.Propagator;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.analytical.EcksteinHechlerPropagator;
import org.orekit.propagation.events.ApsideDetector;
import org.orekit.propagation.events.EclipseDetector;
import org.orekit.propagation.events.ElevationDetector;
import org.orekit.propagation.events.EventDetector;
import org.orekit.propagation.events.FootprintOverlapDetector;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeComponents;
import org.orekit.time.TimeScale;
import org.orekit.time.TimeScalesFactory;
import org.orekit.utils.IERSConventions;
import org.orekit.utils.PVCoordinatesProvider;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.DoubleStream;


import static org.orekit.utils.Constants.*;

public class TestCase {

  public static InputStream getFileFromResourceAsStream(String fileName) {

    // The class loader that loaded the class
    ClassLoader classLoader = TestCase.class.getClassLoader();
    InputStream inputStream = classLoader.getResourceAsStream(fileName);

    // the stream holding the file content
    if (inputStream == null) {
      throw new IllegalArgumentException("file not found! " + fileName);
    } else {
      return inputStream;
    }
  }


  public static void main(String[] args) throws ParseException, IOException {

    double[] altitude = DoubleStream.iterate(400e3, i -> i + 20e3).limit(11).toArray(); // (m) Satellite altitude from 400km to 600km
    File orekitData = new File("PATHTOYOUROREKITDATA\\orekit-data"); //TODO: change your path here
    DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
    manager.addProvider(new DirectoryCrawler(orekitData));

    //region 1- Dates
    Satellite satellite = new Satellite();
    satellite.setUdInitialDate("31-12-2020 22:30:00.000 UTC"); // FIXME: changing this date gives different results with the footprintoverlapdetector
    satellite.setUdFinalDate("02-01-2021 00:00:00.000 UTC");

    // Setting Scenario
    TimeScale timescale = TimeScalesFactory.getUTC();
    SimpleDateFormat sdf =
        new SimpleDateFormat("dd-MM-yyyy HH:mm:ss.SSS Z");

    // Setting Inserted Satellite Object
    Date dateSatelliteInitialDate =
        sdf.parse(satellite.getUdInitialDate());
    Date dateSatelliteFinalDate =
        sdf.parse(satellite.getUdFinalDate());
    AbsoluteDate satelliteInitialDate =
        new AbsoluteDate(dateSatelliteInitialDate, timescale);
    AbsoluteDate satelliteFinalDate =
        new AbsoluteDate(dateSatelliteFinalDate, timescale);
    //endregion

    //region 2- Orbit
    io.testcase.model.Orbit orbit = new io.testcase.model.Orbit();

    //region 2.1- Orbit parameters

    orbit.setPerigeeAltitude(altitude[0]);
    orbit.setApogeeAltitude(altitude[0] + 0.1e3);
    orbit.setAop(FastMath.toRadians(0));
    orbit.setAnomaly(FastMath.toRadians(0));
    orbit.setInc(FastMath.toRadians(90));
    orbit.setRaan(FastMath.toRadians(0));
    orbit.setAnomalyType("True");
    orbit.setA(EIGEN5C_EARTH_EQUATORIAL_RADIUS + 0.5 * (orbit.getPerigeeAltitude() + orbit.getPerigeeAltitude()));
    orbit.setEcc((orbit.getApogeeAltitude() - orbit.getPerigeeAltitude()) / (orbit.getApogeeAltitude() + orbit.getPerigeeAltitude()));
    //endregion

    //region 2.2- Main Body attractor
    Frame earthFrame = FramesFactory.getITRF(IERSConventions.IERS_2010, true);
    OneAxisEllipsoid earth = new OneAxisEllipsoid(
        WGS84_EARTH_EQUATORIAL_RADIUS,
        WGS84_EARTH_FLATTENING,
        earthFrame
    );
    //endregion

    //region 2.3- Attitude
    Frame satelliteFrame = FramesFactory.getGCRF();
    AttitudeProvider pointing = new NadirPointing(satelliteFrame, earth);
    //endregion

    //region 2.4- Setting Keplerian Orbit
    PositionAngle positionAngle = PositionAngle.TRUE;
    Orbit initialOrbit = new KeplerianOrbit(
        orbit.getA(),
        orbit.getEcc(),
        orbit.getInc(),
        orbit.getAop(),
        orbit.getRaan(),
        orbit.getAnomaly(),
        positionAngle,
        satelliteFrame,
        satelliteInitialDate,
        EIGEN5C_EARTH_MU
    );
    //endregion


    //endregion

    //region 3- Mass properties
    double mass = 25.0d;
    //endregion

    //region 4- Propagator
    double stepSize = 60d;

    Propagator propagator = new EcksteinHechlerPropagator(
        initialOrbit,
        pointing,
        mass,
        EIGEN5C_EARTH_EQUATORIAL_RADIUS,
        EIGEN5C_EARTH_MU,
        EIGEN5C_EARTH_C20,
        EIGEN5C_EARTH_C30,
        EIGEN5C_EARTH_C40,
        EIGEN5C_EARTH_C50,
        EIGEN5C_EARTH_C60
    );
    //endregion

    //region 5- Ground Stations
    List<Target> targetList = new ArrayList<>();

    // Fairbanks GST
    Target target1 = new Target();
    target1.setTargetName("Fairbanks");
    target1.setTargetLatitude(FastMath.toRadians(64.8));
    target1.setTargetLongitude(FastMath.toRadians(-147.7));
    target1.setTargetAltitude(135);
    targetList.add(target1);
    //endregion

    //region 6
    for (Target target : targetList) {
      GeodeticPoint geodeticPoint = new GeodeticPoint(target.getTargetLatitude(), target.getTargetLongitude(), target.getTargetAltitude());
      TopocentricFrame frame = new TopocentricFrame(earth, geodeticPoint, target.getTargetName());


      //region 6.1- Access Parameters
      Access access = new Access();
      access.setMaxCheck(60);
      access.setThreshold(0.001);
      access.setMinimumElevation(FastMath.toRadians(5.));
      //endregion

      //region 6.2- Calculation accesses by adding handler and detector to propagator
      StringBuilder sbVisibility = new StringBuilder();
      EventDetector visibility =
          new ElevationDetector(access.getMaxCheck(), access.getThreshold(), frame).
              withConstantElevation(access.getMinimumElevation()).
              withHandler(
                  (s, detector, increasing) -> {
                    sbVisibility.append(
                        " Visibility on " +
                            detector.getTopocentricFrame().getName() +
                            (increasing ? " begins at " : " ends at ") +
                            s.getDate() + "\n"
                    );
                    return Action.CONTINUE;
                  }
              );

      propagator.addEventDetector(visibility);
    }

    //endregion

    //region 7- Eclipses
    PVCoordinatesProvider sun = CelestialBodyFactory.getSun();
    Map<AbsoluteDate, String> earthDayNightInfo = new TreeMap<>();

    EventDetector dayNightEvent = new EclipseDetector(sun, SUN_RADIUS, earth).withUmbra().
        withHandler((s, detector, increasing) -> {
          earthDayNightInfo.put(s.getDate(), (increasing ? "Umbra finishes/Penumbra starts " : "Penumbra finishes/Umbra starts "));
          return Action.CONTINUE;
        });
    EventDetector nightDayEvent = new EclipseDetector(sun, SUN_RADIUS, earth).withPenumbra().
        withHandler((s, detector, increasing) -> {
          earthDayNightInfo.put(s.getDate(), (increasing ? "Penumbra finishes/Light starts " : "Light finishes/Penumbra starts "));
          return Action.CONTINUE;
        });

    propagator.addEventDetector(dayNightEvent);
    propagator.addEventDetector(nightDayEvent);
    //endregion

    //region 8- Land Requirement
    double hyperplaneThickness = 1.0e-10;
    double samplingStep = 100000.; // Number to choose number of splits of the polygon (inversely proportional)
    Map<AbsoluteDate, String> targetLandDetectorInfo = new TreeMap<>(); // Holds information of footprint detector

    List<String> listOfFileNames = List.of("americas.geojson");
    List<FootprintOverlapDetector> listOfDetectors = new ArrayList<>();
    for (String fileName : listOfFileNames) {
      SphericalPolygonsSet sphericalPolygonsSet = null;
      FeatureCollection featureCollection = new ObjectMapper().readValue(getFileFromResourceAsStream(fileName), FeatureCollection.class);
      //this represents all continents
      for (Feature feature : featureCollection.getFeatures()) {
        if (feature.getGeometry() instanceof Polygon) {
          // this represents all vertices of the current continent
          List<GeodeticPoint> geodeticPoints = new ArrayList<>();
          for (List<LngLatAlt> coordinates : ((Polygon) feature.getGeometry()).getCoordinates()) {
            for (LngLatAlt lngLatAlt : coordinates) {
              geodeticPoints.add(new GeodeticPoint(FastMath.toRadians(lngLatAlt.getLatitude()), FastMath.toRadians(lngLatAlt.getLongitude()), 0));
            }
          }
          GeodeticPoint[] geodeticPointArray = geodeticPoints.toArray(new GeodeticPoint[]{});
          sphericalPolygonsSet = EllipsoidTessellator.buildSimpleZone(hyperplaneThickness, geodeticPointArray);
        }
      }

      //region
      double focalLength = 0.297; // (m) Focal length
      double pixelsAlongTrackTIR = 512; // Number of pixels along track
      double pixelsAcrossTrackTIR = 640; // Number of pixels across track
      double pixelPitchTIR = 15.0e-6; // (m) Pixel size assuming square pixel
      //endregion

      double acrossTrackFieldOfView = 2 * Math.atan(pixelPitchTIR * pixelsAcrossTrackTIR / (2 * focalLength)); // (rad) Field of View (FoV) across track
      double alongTrackFieldOfView = 2 * Math.atan(pixelPitchTIR * pixelsAlongTrackTIR / (2 * focalLength)); // (rad) Field of View (FoV) along track
      DoubleDihedraFieldOfView imagerFieldOfViewArray = new DoubleDihedraFieldOfView(Vector3D.PLUS_K, Vector3D.PLUS_I, alongTrackFieldOfView / 2, Vector3D.PLUS_J, acrossTrackFieldOfView / 2, 0);

      FootprintOverlapDetector targetLandDetector = new FootprintOverlapDetector(imagerFieldOfViewArray, earth, sphericalPolygonsSet, samplingStep).withHandler((s, detector, increasing) -> {
        targetLandDetectorInfo.put(s.getDate(), (increasing ? "Start seeing land at " : "Stops seeing land at "));
        return Action.CONTINUE;
      });
      propagator.addEventDetector(targetLandDetector);
      listOfDetectors.add(targetLandDetector);
    }
    //endregion

    //region 9- Apside Detector
    List<LngLatAlt> listOfSubPoints = new ArrayList<>();

    AtomicInteger numberOfOrbits = new AtomicInteger(0);
    ApsideDetector apsideDetector = new ApsideDetector(initialOrbit).withHandler((s, detector, increased) -> {
      if (increased) {
        numberOfOrbits.incrementAndGet();
      }

      return Action.CONTINUE;
    });
    propagator.addEventDetector(apsideDetector);
    //endregion

    //region 10- Propagation Loop
    AbsoluteDate propagationDate;
    SpacecraftState currentState;
    for (
        propagationDate = satelliteInitialDate;
        propagationDate.compareTo(satelliteFinalDate) <= 0;
        propagationDate = propagationDate.shiftedBy(stepSize)

    ) {
      currentState = propagator.propagate(propagationDate);
      GeodeticPoint satelliteSubPoint = earth.transform(currentState.getPVCoordinates().getPosition(), satelliteFrame, currentState.getDate());
      listOfSubPoints.add(new LngLatAlt(FastMath.toDegrees(satelliteSubPoint.getLatitude()), FastMath.toDegrees(satelliteSubPoint.getLongitude()), 0));
    }
    targetLandDetectorInfo.forEach((key, value) -> System.out.println(value + key));

    //endregion

  }
}
